//
//  BonanzaAppDelegate.m
//  Bonanza
//
//  Created by 永田 央 on 12/03/24.
//  Copyright 2012 __MyCompanyName__. All rights reserved.
//

#import "BonanzaAppDelegate.h"
#include <unistd.h>

const char *komaNames[] = {
	NULL, "FU", "KY", "KE", "GI", "KI", "KA", "HI", "OU", NULL,
	NULL, "TO", "NY", "NK", "NG", NULL, "UM", "RY" };

#define kBoardXSize 410
#define kBoardYSize 454
#define kBoardXMargin 5
#define kBoardYMargin 5
#define kKomaXSize 43
#define kKomaYSize 48
#define kKomaXOffset 11
#define kKomaYOffset 11
#define kDaiXSize 135
#define kDaiYSize 217
#define kDaiXMargin 5
#define kDaiXPadding 3
#define kDaiYPadding 11
#define kKomaXOrigin (kDaiXMargin + kDaiXSize + kBoardXMargin + kKomaXOffset)
#define kKomaYOrigin (kBoardYMargin + kKomaYOffset)

enum {
	kErrorUnsanePosition = -1,
	kErrorUnsaneKomaCode = -2,
	kErrorNoKomaIsThere = -3,
};

int
komaStringToCode(const char *komaString)
{
	int i;
	char upperKoma[3];
	upperKoma[0] = toupper(komaString[0]);
	upperKoma[1] = toupper(komaString[1]);
	for (i = 1; i < 18; i++) {
		if (komaNames[i] != NULL && strncmp(upperKoma, komaNames[i], 2) == 0)
			return i;
	}
	return 0;  /*  Invalid name  */
}

const char *
komaCodeToString(int code)
{
	if (code > 0 && code < 18)
		return komaNames[code];
	else return NULL;
}

NSString *gKomaImage = @"komaImage",
	*gBanImage = @"banImage",
	*gGridImage = @"gridImage",
	*gBlackIsHuman = @"blackIsHuman",
	*gWhiteIsHuman = @"whiteIsHuman",
	*gHandicap = @"handicap",
	*gAvailableTime1 = @"availableTime1",
	*gAvailableTime2 = @"availableTime2";

NSImage *gBlackFront, *gBlackBack, *gWhiteFront, *gWhiteBack;

@implementation BonanzaAppDelegate

#pragma mark ====== Loading Images ======

- (void)loadKomaImages:(int)tag
{
	NSString *identifier, *respath;
	int i, j;
	static char *names[] = {
		NULL, "fu", "kyo", "kei", "gin", "kin", "kaku", "hi", "ou", NULL,
		NULL, "to", "nkyo", "nkei", "ngin", NULL, "uma", "ryu" };

	switch (tag) {
		case 1: identifier = @"kinki_tiger"; break;
		case 2: identifier = @"ryoko_regular"; break;
		case 3: identifier = @"ryoko_tiger"; break;
		case 4: identifier = @"kinki_abbr"; break;
		default: identifier = @"kinki_regular"; break;
	}
	respath = [[NSBundle mainBundle] pathForResource:identifier ofType:nil];
	for (i = 0; i < 2; i++) {
		for (j = 0; j < 18; j++) {
			[images[i][j] release];
			if (names[j] == NULL) {
				images[i][j] = nil;
				continue;
			} else {
				images[i][j] = [[NSImage alloc] initWithContentsOfFile:[NSString stringWithFormat:@"%@/%c%s.png", respath, (i == 0 ? 'S' : 'G'), names[j]]];
			}
		}
	}
	
	for (i = 0; i < 2; i++) {
		int base = (i == 0 ? 100 : 200);
		for (j = 1; j <= 7; j++) {
			id view = [[window contentView] viewWithTag:base + j];
			[view setHidden:YES];
			[view setImage:images[i][j]];
			view = [[window contentView] viewWithTag:base + 10 + j];
			[view setHidden:YES];
			[view setStringValue:@""];
		}
	}
	if (gBlackFront == nil)
		gBlackFront = [[NSImage imageNamed:@"black_front.png"] retain];
	if (gBlackBack == nil)
		gBlackBack = [[NSImage imageNamed:@"black_back.png"] retain];
	if (gWhiteFront == nil)
		gWhiteFront = [[NSImage imageNamed:@"white_front.png"] retain];
	if (gWhiteBack == nil)
		gWhiteBack = [[NSImage imageNamed:@"white_back.png"] retain];
	
	[shogiView setNeedsDisplay:YES];
}

- (void)loadBanImage:(int)tag
{
	NSString *identifier, *respath;
	switch (tag) {
		case 1: identifier = @"kaya_b"; break;
		case 2: identifier = @"kaya_c"; break;
		case 3: identifier = @"kaya_d"; break;
		case 4: identifier = @"gohan"; break;
		case 5: identifier = @"muji"; break;
		case 6: identifier = @"stripe"; break;
		case 7: identifier = @"oritatami"; break;
		case 8: identifier = @"dirty"; break;
		case 9: identifier = @"paper"; break;
		default: identifier = @"kaya_a"; break;
	}
	respath = [[NSBundle mainBundle] pathForResource:@"board" ofType:nil];
	[banImage release];
	banImage = [[NSImage alloc] initWithContentsOfFile:[NSString stringWithFormat:@"%@/ban_%@.png", respath, identifier]];
	[shogiView setNeedsDisplay:YES];
}

- (void)loadGridImage:(int)tag
{
	NSString *identifier, *respath;
	switch (tag) {
		case 1: identifier = @"dot_xy"; break;
		case 2: identifier = @"nodot"; break;
		case 3: identifier = @"nodot_xy"; break;
		case 4: identifier = @"handwriting"; break;
		default: identifier = @"dot"; break;
	}
	respath = [[NSBundle mainBundle] pathForResource:@"grid" ofType:nil];
	[gridImage release];
	gridImage = [[NSImage alloc] initWithContentsOfFile:[NSString stringWithFormat:@"%@/masu_%@.png", respath, identifier]];
	[shogiView setNeedsDisplay:YES];
}

#pragma mark ====== Game ======

- (void)setUpNewGame
{
	NSUserDefaults *def = [NSUserDefaults standardUserDefaults];
	int i;
	signed char init_board[81] = {
		-2, -3, -4, -5, -8, -5, -4, -3, -2,
		0, -6,  0,  0,  0,  0,  0, -7,  0,
		-1, -1, -1, -1, -1, -1, -1, -1, -1,
		0,  0,  0,  0,  0,  0,  0,  0,  0,
		0,  0,  0,  0,  0,  0,  0,  0,  0,
		0,  0,  0,  0,  0,  0,  0,  0,  0,
		1,  1,  1,  1,  1,  1,  1,  1,  1,
		0,  7,  0,  0,  0,  0,  0,  6,  0,
		2,  3,  4,  5,  8,  5,  4,  3,  2
	};
	
	[self stopBonanza];
	
	/*  Initialize board  */
	memmove(board, init_board, 81);
	for (i = 0; i < 8; i++)
		mochigoma[0][i] = mochigoma[1][i] = 0;
	memset(&moving, 0, sizeof(moving));
	
	/*  Handicapped game  */
	handicap = [[def objectForKey:gHandicap] intValue];
	if (handicapSpec != nil) {
		[handicapSpec release];
		handicapSpec = nil;
	}
	switch (handicap) {
		case 1:  /*  Drop left Kyo  */
			board[0] = 0;
			handicapSpec = @"99KY";
			break;
		case 2:  /*  Drop Kaku  */
			board[10] = 0;
			handicapSpec = @"88KA";
			break;
		case 3:  /*  Drop Hisha  */
			board[16] = 0;
			handicapSpec = @"28HI";
			break;
		case 4:  /*  Drop Kaku-Kyo  */
			board[8] = board[10] = 0;
			handicapSpec = @"19KY88KA";
			break;
		case 5:  /*  Drop Hisha-Kyo  */
			board[0] = board[16] = 0;
			handicapSpec = @"99KY28HI";
			break;
		case 6:  /*  Drop 2-mai  */
			board[10] = board[16] = 0;
			handicapSpec = @"88KA28HI";
			break;
		case 7:  /*  Drop 4-mai  */
			board[0] = board[8] = board[10] = board[16] = 0;
			handicapSpec = @"88KA28HI99KY19KY";
			break;
		case 8:  /*  Drop 6-mai  */
			board[0] = board[1] = board[7] = board[8] = board[10] = board[16] = 0;
			handicapSpec = @"88KA28HI99KY19KY89KE29KE";
			break;
		case 9:  /*  Drop 8-mai  */
			board[0] = board[1] = board[2] = board[6] = board[7] = board[8] = 0;
			board[10] = board[16] = 0;
			handicapSpec = @"88KA28HI99KY19KY89KE29KE79GI39GI";
			break;
		case 10: /*  Hisha/Kaku/Kyo/Kei only  */
			board[2] = board[3] = board[5] = board[6] = 0;
			handicapSpec = @"99GI39GI69KI49KI";
			break;
		case 11: /*  Hisha/Kaku/Kyo only  */
			board[1] = board[2] = board[3] = 0;
			board[5] = board[6] = board[7] = 0;
			handicapSpec = @"99KE29KE79GI39GI69KI49KI";
			break;
		case 12: /*  Hisha/Kaku only  */
			board[0] = board[1] = board[2] = board[3] = 0;
			board[5] = board[6] = board[7] = board[8] = 0;
			handicapSpec = @"99KY19KY89KE29KE79GI39GI69KI49KI";
			break;
		case 13: /*  Drop 10-mai  */
			board[0] = board[1] = board[2] = board[3] = board[5] = board[6] = 0;
			board[7] = board[8] = board[10] = board[16] = 0;
			handicapSpec = @"88KA28HI99KY19KY89KE29KE79GI39GI69KI49KI";
			break;
		default: /*  No handicap  */
			handicapSpec = @"";
			break;
	}
	if (handicap != 0)
		whiteMovesFirst = YES;
	else whiteMovesFirst = NO;
	teban = -1;
	status = 0;
	{
		NSString *yourTurn = NSLocalizedString(@"Your Turn", @"Your Turn");
		NSString *pondering = NSLocalizedString(@"Pondering...", @"Pondering...");
		isHuman[0] = ([[def objectForKey:gBlackIsHuman] intValue] != 0);
		isHuman[1] = ([[def objectForKey:gWhiteIsHuman] intValue] != 0);
		[blackTebanInfo setStringValue:(isHuman[0] ? yourTurn : pondering)];
		[whiteTebanInfo setStringValue:(isHuman[1] ? yourTurn : pondering)];
		[blackTebanInfo setHidden:YES];
		[whiteTebanInfo setHidden:YES];
	}

	[bonMoves removeAllObjects];
	bonMoveIndex = 0;
	
	elapsedTime[0] = elapsedTime[1] = 0;

	[shogiView setNeedsDisplay:YES];
	[self updateTebanInfo];
}

static int
setOneMovablePosition(unsigned char *pos, const signed char *board, int x, int y, int sign)
{
	int k;
	if (x <= 0 || x >= 10 || y <= 0 || y >= 10)
		return -1;  /*  Out of range  */
	k = x + y * 9 - 10;
	if (board[k] == 0) {
		pos[k] = 1;
		return 0;  /*  OK  */
	} else if (board[k] * sign > 0) {
		/*  My koma is already there  */
		return -2;
	} else {
		/*  Opponent's koma is there  */
		pos[k] = 1;  /*  Movable  */
		return -3;   /*  But cannot move any further  */
	}
}

/*  Set the movable positions in movable[81] array.  */
/*  Note: movable[] is _not_ zero-cleared in this function.  */
/*  For checking 'legal' positions, i.e. movement that does not result in 'check' of
	the player's ou, use setAllowablePositions:forKomaAtX:y:.  */
- (int)setMovablePositions:(unsigned char *)movable forKomaAtX:(int)x y:(int)y
{
	int koma, sign, k, xx, yy, i;
	if (x <= 0 || x >= 10 || y <= 0 || y >= 10)
		return kErrorUnsanePosition;
	k = x + y * 9 - 10;
	koma = board[k];
	if (koma == 0)
		return kErrorNoKomaIsThere;
	if (koma > 0)
		sign = 1;
	else {
		koma = -koma;
		sign = -1;
	}
	switch (koma) {
		case 1:  /* Fu */
			setOneMovablePosition(movable, board, x, y - sign, sign);
			break;
		case 2:  /* Kyo */
			yy = y;
			do {
				yy -= sign;
			} while (setOneMovablePosition(movable, board, x, yy, sign) == 0);
			break;
		case 3:  /* Kei */
			setOneMovablePosition(movable, board, x + 1, y - sign * 2, sign);
			setOneMovablePosition(movable, board, x - 1, y - sign * 2, sign);
			break;
		case 4:  /* Gin */ {
			static const signed char gin_ofs[] = {-1, 1, 0, 1, 1, 1, -1, -1, 1, -1};
			for (i = 0; i < 10; i += 2)
				setOneMovablePosition(movable, board, x + gin_ofs[i], y - sign * gin_ofs[i + 1], sign);
			break;
		}
		case 5:  /* Kin */
		case 11: /* To */
		case 12: /* Nari-Kyo */
		case 13: /* Nari-Kei */
		case 14: /* Nari-Gin */ {
			static const signed char kin_ofs[] = {-1, 1, 0, 1, 1, 1, -1, 0, 1, 0, 0, -1};
			for (i = 0; i < 12; i += 2)
				setOneMovablePosition(movable, board, x + kin_ofs[i], y - sign * kin_ofs[i + 1], sign);
			break;
		}
		default:
			if (koma == 8 || koma == 16 || koma == 17) {
				/* Ou, Uma, Ryu */
				static const signed char ou_ofs[] = {-1, 1, 0, 1, 1, 1, -1, 0, 1, 0, -1, -1, 0, -1, 1, -1};
				for (i = 0; i < 16; i += 2)
					setOneMovablePosition(movable, board, x + ou_ofs[i], y - sign * ou_ofs[i + 1], sign);
			}
			if (koma == 6 || koma == 16) {
				/*  Kaku or Uma  */
				static const signed char kaku_ofs[] = {1, 1, 1, -1, -1, 1, -1, -1};
				for (i = 0; i < 8; i += 2) {
					xx = x;
					yy = y;
					do {
						xx += kaku_ofs[i];
						yy += kaku_ofs[i + 1];
					} while (setOneMovablePosition(movable, board, xx, yy, sign) == 0);
				}
			} else if (koma == 7 || koma == 17) {
				/*  Hisha or Ryu  */
				static const signed char hisha_ofs[] = {0, 1, 0, -1, 1, 0, -1, 0};
				for (i = 0; i < 8; i += 2) {
					xx = x;
					yy = y;
					do {
						xx += hisha_ofs[i];
						yy += hisha_ofs[i + 1];
					} while (setOneMovablePosition(movable, board, xx, yy, sign) == 0);
				}
			}
			break;
	}
	return 0;
}

/*  Is the player's ou is checked (ou-te)  */
- (BOOL)isChecked:(unsigned char)bw
{
	int x, y, k, op;
	unsigned char movable[81];
	memset(movable, 0, 81);
	op = -1;
	for (x = 1; x <= 9; x++) {
		for (y = 1; y <= 9; y++) {
			k = board[x + y * 9 - 10];
			if ((bw == 0 && k < 0) || (bw == 1 && k > 0))
				[self setMovablePositions:movable forKomaAtX:x y:y];
			else if ((bw == 0 && k == 8) || (bw == 1 && k == -8)) {
				/*  Ou position  */
				op = x + y * 9 - 10;
			}
		}
	}
	if (op >= 0 && movable[op] != 0)
		return YES;
	else return NO;
}

/*  Set the allowable positions for Koma at (x,y) to be able to move in movable[81] array.  */
/*    0: not movable  */
/*    1: movable and legal  */
/*    2: movable but illegal  */
/*  Note: unlike setMovablePositions:forKomaAtX:y:, this method _does_ clear the
 movable[] array.  */
- (int)setAllowablePositions:(unsigned char *)movable forKomaAtX:(int)x y:(int)y
{
	int n, k, kk, xx, yy, koma1, koma2;
	unsigned char bw;
	memset(movable, 0, 81);
	n = [self setMovablePositions:movable forKomaAtX:x y:y];
	if (n != 0)
		return n;
	/*  Make sure that the player's ou is not checked  */
	k = x + y * 9 - 10;
	koma1 = board[k];
	bw = (koma1 < 0);
	board[k] = 0;
	for (xx = 1; xx <= 9; xx++) {
		for (yy = 1; yy <= 9; yy++) {
			kk = xx + yy * 9 - 10;
			if (movable[kk] == 0)
				continue;
			koma2 = board[kk];
			board[kk] = koma1;
			if ([self isChecked:bw])
				movable[kk] = 2;
			board[kk] = koma2;
		}
	}
	board[k] = koma1;
	return 0;
}

/*  Is the player's ou mated?  */
- (BOOL)isMated:(unsigned char)bw
{
	int x, y, sign;
	if (![self isChecked:bw])
		return NO;
	sign = (bw == 0 ? 1 : -1);
	for (x = 1; x <= 9; x++) {
		for (y = 1; y <= 9; y++) {
			int k, koma1;
			koma1 = board[x + y * 9 - 10];
			if (koma1 * sign > 0) {
				unsigned char movable[81];
				[self setAllowablePositions:movable forKomaAtX:x y:y];
				for (k = 0; k < 81; k++) {
					if (movable[k] == 1)
						return NO; /*  There is a possible move to avoid ou-te  */
				}
			}
		}
	}
	return YES;  /*  No move to avoid ou-te  */
}

/*  Set the droppable positions in droppable[81] array. For white's turn, give a negative koma. */
/*  Note: unlike setMovablePositions:forKomaAtX:y:, this function clears non-droppable positions. */
- (int)setDroppablePositions:(unsigned char *)droppable forKoma:(int)koma
{
	int i, n1, n2, sign;
	char fu[9];
	memset(droppable, 0, 81);
	memset(fu, 0, 9);
	sign = (koma > 0 ? 1 : -1);
	if (koma == 1 || koma == -1) {
		/* Check for ni-fu */
		for (n1 = 0; n1 < 9; n1++) {
			fu[n1] = 0;
			for (n2 = 0; n2 < 9; n2++) {
				if (board[n1 + n2 * 9] == koma) {
					fu[n1] = 1;
					break;
				}
			}
		}
	}
	n1 = 0;
	n2 = 80;
	if (koma == 1 || koma == 2) {
		/* Black's fu or kyo */
		n1 = 9;
	} else if (koma == 3) {
		/* Black's kei */
		n1 = 18;
	} else if (koma == -1 || koma == -2) {
		/* White's fu or kyo */
		n2 = 71;
	} else if (koma == -3) {
		/* White's kei */
		n2 = 62;
	}
	if (koma == 1 || koma == -1) {
		for (i = n1; i <= n2; i++) {
			if (fu[i % 9] == 1)
				continue; /* ni-fu */
			if (board[i] == 0) {
				droppable[i] = 1;
				if (board[i - sign * 9] == -sign * 8) {
					/*  Opponent's ou  */
					/*  Check for mate by fu drop  */
					board[i] = sign;
					if ([self isMated:(sign > 0)])
						droppable[i] = 0;  /*  Mate by fu drop: illegal  */
					board[i] = 0;
				}
			}
		}
	} else {
		for (i = n1; i <= n2; i++) {
			if (board[i] == 0)
				droppable[i] = 1;
		}
	}
	return 0;
}

- (BOOL)canUndo
{
	if (teban == -1 || bonMoveIndex == 0)
		return NO;
	if (isHuman[0] == 0 && isHuman[1] == 0)
		return NO;  /*  Undo is only for human  */
	if (bonMoveIndex == 1 && isHuman[whiteMovesFirst] == 0)
		return NO;  /*  Only the computer made the first move: nothing to undo  */
	return YES;
}

#pragma mark ====== Communicate with Bonanza ======

- (void)startBonanza
{
	NSString *execPath;

	if (bonTask != nil || pipeToBon != nil || pipeFromBon != nil)
		[self stopBonanza];

	/*  Create new pipes  */
	pipeToBon = [[NSPipe alloc] init];
	pipeFromBon = [[NSPipe alloc] init];
	
	/*  Prepare launching bonanza  */
	execPath = [[NSBundle mainBundle] pathForAuxiliaryExecutable:@"bonanza"];
	bonTask = [[NSTask alloc] init];
	[bonTask setLaunchPath:execPath];
	[bonTask setStandardInput:pipeToBon];
	[bonTask setStandardOutput:pipeFromBon];
	[bonTask setCurrentDirectoryPath:bonPath];
	
	[[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(readDataFromBon:) name:NSFileHandleDataAvailableNotification object:[pipeFromBon fileHandleForReading]];
	[[pipeFromBon fileHandleForReading] waitForDataInBackgroundAndNotify];
	
	[bonLastUnfinishedMessage release];
	bonLastUnfinishedMessage = @"";
	
	[bonTask launch];
}

- (void)stopBonanza
{
	/*  Destroy previous task and pipe if present  */
	if (bonTask != nil) {
		if ([bonTask isRunning])
			[bonTask terminate];
		[bonTask release];
		bonTask = nil;
	}
	if (pipeToBon != nil)
		[pipeToBon release];
	if (pipeFromBon != nil) {
		[[NSNotificationCenter defaultCenter] removeObserver:self name:NSFileHandleDataAvailableNotification object:[pipeFromBon fileHandleForReading]];
		[pipeFromBon release];
	}
	pipeToBon = pipeFromBon = nil;
}

- (void)writeStringToBon:(NSString *)str
{
	NSData *data = [str dataUsingEncoding:NSUTF8StringEncoding];
	[[pipeToBon fileHandleForWriting] writeData:data];
//	printf(">>>>>> %s", [str UTF8String]);
}

- (void)readDataFromBon:(NSNotification *)notification
{
	NSFileHandle *fileHandle = (NSFileHandle *)[notification object];
	NSData *data = [fileHandle availableData];
	if ([data length] > 0) {
		int i, j;
		NSArray *ary;
		NSString *str = [[[NSString alloc] initWithData:data encoding:NSUTF8StringEncoding] autorelease];
		if ([str hasSuffix:@"Black Mated> "] || [str hasSuffix:@"White Mated> "]) {
			/*  End of game --- process the last 'unfinished' line */
			str = [str stringByAppendingString:@"\n"];
		}
		ary = [[bonLastUnfinishedMessage stringByAppendingString:str] componentsSeparatedByString:@"\n"];
		[bonLastUnfinishedMessage release];
		bonLastUnfinishedMessage = [[ary lastObject] retain];
		/*  Append lines matching "Black .*>" or "White .*>"  */
		for (i = 0; i < [ary count] - 1; i++) {
			NSString *s = [ary objectAtIndex:i];
			const char *cs = [s UTF8String];
			BonMoveInfo info;
			memset(&info, 0, sizeof(info));
			if (cs[0] == 7)  /*  Bell  */
				cs++;
		//	NSLog(@"%s", cs);
			if (strncmp(cs, "Black ", 6) == 0)
				info.bw = 0;
			else if (strncmp(cs, "White ", 6) == 0)
				info.bw = 1;
			else continue;
			cs += 6;
			if (strncmp(cs, "Mated>", 6) == 0) {
				info.bw = 1 - info.bw;
				info.koma = 98;
			} else {
				/*  "tekazu>" */
				if (!isdigit(cs[0]))
					continue;
				info.moves = strtol(cs, (char **)&cs, 0);
				if (*cs++ != '>')
					continue;
				if (*cs++ != ' ')
					continue;
				if (strncmp(cs, "resign", 6) == 0) {
					info.koma = 99;
				} else {
					/*  Move command  */
					if (!isdigit(cs[0]) || !isdigit(cs[1]) || !isdigit(cs[2]) || !isdigit(cs[3]))
						continue;
					info.pos[0] = cs[0] - '0';
					info.pos[1] = cs[1] - '0';
					info.pos[2] = cs[2] - '0';
					info.pos[3] = cs[3] - '0';
					if (cs[4] < 'A' || cs[4] > 'Z' || cs[5] < 'A' || cs[5] > 'Z')
						continue;
					else {
						char komaName[3];
						komaName[0] = cs[4];
						komaName[1] = cs[5];
						komaName[2] = 0;
						info.koma = komaStringToCode(komaName);
						if (info.koma == 0)
							continue;
					}
					cs += 6;
					/*  Statistics  */
					if (strncmp(cs, " '(", 3) == 0) {
						char *p = strstr(cs, "elapsed: ");
						if (p == NULL) {
							strncpy(info.mes, cs, 31);
							info.mes[31] = 0;
						} else {
							int len2 = p - cs;
							if (len2 > 31)
								len2 = 31;
							strncpy(info.mes, cs, len2);
							info.mes[len2] = 0;
							cs = p + 8;
							p = strchr(cs, 'b');
							if (p != NULL)
								info.b_elapse = strtol(p + 1, NULL, 0);
							else info.b_elapse = -1;
							p = strchr(cs, 'w');
							if (p != NULL)
								info.w_elapse = strtol(p + 1, NULL, 0);
							else info.w_elapse = -1;
						}
					}
				}
			}
			if (whiteMovesFirst) {
				info.bw = 1 - info.bw;
				for (j = 0; j < 4; j++) {
					if (info.pos[j] > 0)
						info.pos[j] = 10 - info.pos[j];
				}
				j = info.w_elapse;
				info.w_elapse = info.b_elapse;
				info.b_elapse = j;
			}
			/*  Captured koma  */
			j = board[info.pos[2] + info.pos[3] * 9 - 10];
			if (j < 0)
				j = -j;
			info.capture = j;
			j = board[info.pos[0] + info.pos[1] * 9 - 10];
			if (j < 0)
				j = -j;
			if (j + 10 == info.koma)
				info.promote = 1;
			/*  Handle move  */
			info.moveTime = elapsedTime[teban] - lastMoveTime;
			[self stopTimer];
			[self makeMove:&info withAnimation:YES];
			[shogiView display];
			/*  If the next move is also bonanza then ask for next move  */
			if (teban >= 0 && isHuman[teban] == 0)
				[self writeStringToBon:@"move\n"];
			[self startTimer];
		}
	}
	[fileHandle waitForDataInBackgroundAndNotify];
}

#pragma mark ====== Display ======

- (void)setNeedsDisplayBoard
{
	[shogiView setNeedsDisplay:YES];
	[blackTebanInfo setNeedsDisplay:YES];
	[whiteTebanInfo setNeedsDisplay:YES];
}

static NSRect
mochigomaRect(int bw, int koma, int index, int count, int whiteIsThisSide)
{
	int x, y, k;
	BOOL thisSide;
	if (count == 0)
		return NSZeroRect;
	if (koma == 1) {
		y = kDaiYPadding + kKomaYSize * 3;
		if (count == 1)
			x = kDaiXSize / 2 - kKomaXSize / 2;
		else if (count == 2) {
			k = (kDaiXSize - kDaiXPadding * 2 - kKomaXSize * 2) / 3;
			x = kDaiXPadding + k + (kKomaXSize + k) * index;
		} else {
			k = (kDaiXSize - kDaiXPadding * 2 - kKomaXSize) / (count > 9 ? 8 : count - 1);
			x = kDaiXPadding + k * (index % 9);
			if (index >= 9) {
				x += k / 2;
				y += 6;
			}
		}
	} else {
		y = kDaiYPadding + kKomaYSize * ((7 - koma) / 2);
		k = (kDaiXSize - kDaiXPadding * 2) / 2;
		if (count == 1)
			x = k / 2 - kKomaXSize / 2;
		else
			x = kDaiXPadding + index * ((k - kKomaXSize) / (count - 1));
		if (koma % 2 == 1)
			x += k + kDaiXPadding;
	}
	
	thisSide = (whiteIsThisSide ? bw == 1 : bw == 0);
	if (thisSide) {
		x += kDaiXMargin + kDaiXSize + kBoardXMargin * 2 + kBoardXSize;
		y += kBoardYMargin;
		return NSMakeRect(x, y, kKomaXSize, kKomaYSize);
	} else {
		x = kDaiXSize - x - kKomaXSize;
		y = kDaiYSize - y - kKomaYSize;
		x += kDaiXMargin;
		y += kBoardYMargin + kBoardYSize - kDaiYSize;
		return NSMakeRect(x, y, kKomaXSize, kKomaYSize);
	}
}

- (void)displayBoardInView:(NSView *)boardView
{
	int i, j, n, k;
	NSRect boardRect, blackDaiRect, whiteDaiRect, dstRect;
	id image;
	static NSColor *sRed, *sTransparentRed, *sGreen;
	static NSImage *sTextBackgroundImage;
	const BonMoveInfo *iptr;

	if (sRed == nil)
		sRed = [[NSColor colorWithDeviceRed:1.0 green:0.0 blue:0.0 alpha:1.0] retain];
	if (sTransparentRed == nil)
		sTransparentRed = [[NSColor colorWithDeviceRed:1.0 green:0.0 blue:0.0 alpha:0.25] retain];
	if (sGreen == nil)
		sGreen = [[NSColor colorWithDeviceRed:0.0 green:0.5 blue:0.0 alpha:0.45] retain];
	if (sTextBackgroundImage == nil)
		sTextBackgroundImage = [[NSImage alloc] initWithContentsOfFile:[[NSBundle mainBundle] pathForResource:@"textBackground" ofType:@"png"]];
	
	if (banImage == nil || gridImage == nil)
		return;  /*  Not ready yet  */
	
	if (teban >= 0 && bonMoveIndex > 0) {
		iptr = (const BonMoveInfo *)[[bonMoves objectAtIndex:bonMoveIndex - 1] bytes];
	} else iptr = NULL;

	[boardView lockFocus];
	
	/*  Draw board  */
	boardRect.origin = NSMakePoint(kDaiXMargin + kDaiXSize + kBoardXMargin, kBoardYMargin);
	boardRect.size = NSMakeSize(kBoardXSize, kBoardYSize);
	[banImage drawInRect:boardRect fromRect:NSZeroRect operation:NSCompositeSourceOver fraction:1.0];
	
	/*  Draw grid  */
	[gridImage drawInRect:boardRect fromRect:NSZeroRect operation:NSCompositeSourceOver fraction:1.0];
	
	/*  Draw koma-dai  */
	blackDaiRect.origin = NSMakePoint(kDaiXMargin + kDaiXSize + kBoardXMargin * 2 + kBoardXSize, kBoardYMargin);
	blackDaiRect.size = NSMakeSize(kDaiXSize, kDaiYSize);
	[banImage drawInRect:blackDaiRect fromRect:NSMakeRect(kBoardXSize - kDaiXSize, kBoardYSize - kDaiYSize, kDaiXSize, kDaiYSize) operation:NSCompositeSourceOver fraction:1.0];
	dstRect.origin.x = blackDaiRect.origin.x + kKomaXSize * 0.1;
	dstRect.origin.y = blackDaiRect.origin.y + kDaiYSize + kKomaYSize * 0.1;
	dstRect.size = NSMakeSize(kKomaXSize * 0.6, kKomaYSize * 0.6);
	[(whiteIsThisSide ? gWhiteFront : gBlackFront) drawInRect:dstRect fromRect:NSZeroRect operation:NSCompositeSourceOver fraction:1.0];
	
	whiteDaiRect.origin = NSMakePoint(kDaiXMargin, kBoardYMargin + kBoardYSize - kDaiYSize);
	whiteDaiRect.size = NSMakeSize(kDaiXSize, kDaiYSize);
	[banImage drawInRect:whiteDaiRect fromRect:NSMakeRect(0, 0, kDaiXSize, kDaiYSize) operation:NSCompositeSourceOver fraction:1.0];
	dstRect.origin.x = whiteDaiRect.origin.x + kDaiXSize - kKomaXSize * 0.7;
	dstRect.origin.y = whiteDaiRect.origin.y - kKomaYSize * 0.7;
	dstRect.size = NSMakeSize(kKomaXSize * 0.6, kKomaYSize * 0.6);
	[(whiteIsThisSide ? gBlackBack : gWhiteBack) drawInRect:dstRect fromRect:NSZeroRect operation:NSCompositeSourceOver fraction:1.0];
	
	/*  Draw koma on board  */
	for (i = 0; i < 9; i++) {
		for (j = 0; j < 9; j++) {
			unsigned char focus;
			dstRect.origin = NSMakePoint(boardRect.origin.x + kKomaXOffset + (8 - i) * kKomaXSize, boardRect.origin.y + kKomaYOffset + (8 - j) * kKomaYSize);
			dstRect.size = NSMakeSize(kKomaXSize, kKomaYSize);
			n = i + 9 * j;
			if (iptr != NULL && iptr->pos[2] == i + 1 && iptr->pos[3] == j + 1) {
				[sGreen set];
				NSRectFillUsingOperation(NSInsetRect(dstRect, 1.0, 1.0), NSCompositeSourceOver);
			}
			if (whiteIsThisSide) {
				k = -board[80 - n];
				focus = board_focus[80 - n];
			} else {
				k = board[n];
				focus = board_focus[n];
			}
			if (k != 0) {
				if (k < 0)
					image = images[1][-k];
				else
					image = images[0][k];
				if (focus == 3) {
					[sRed set];
					NSRectFillUsingOperation(NSInsetRect(dstRect, 1.0, 1.0), NSCompositeSourceOver);
				} else {
					[image drawInRect:dstRect fromRect:NSZeroRect operation:NSCompositeSourceOver fraction:(focus == 2 ? 0.5 : 1.0)];
				}
			}
			if (focus == 1) {
				[sTransparentRed set];
				NSRectFillUsingOperation(dstRect, NSCompositeSourceOver);
			}
		}
	}
	
	/*  Draw koma in hand  */
	if (moving.koma > 0 && moving.pos[0] == 0 && moving.pos[1] == 0) {
		dstRect = mochigomaRect(moving.bw, moving.koma, draggingMochigomaIndex, mochigoma[moving.bw][moving.koma - 1], whiteIsThisSide);
		[sRed set];
		NSRectFillUsingOperation(NSInsetRect(dstRect, 1.0, 1.0), NSCompositeSourceOver);		
	}
	
	for (i = 0; i < 2; i++) {
		for (j = 0; j < 7; j++) {
			image = images[i][j + 1];
			n = mochigoma[i][j];
			for (k = n - 1; k >= 0; k--) {
				float fraction = 1.0;
				dstRect = mochigomaRect(i, j + 1, k, n, whiteIsThisSide);
				if (moving.koma == j + 1 && moving.pos[0] == 0 && moving.pos[1] == 0 && moving.bw == i && k == draggingMochigomaIndex) {
					
					[sTransparentRed set];
					NSRectFillUsingOperation(NSInsetRect(dstRect, 1.0, 1.0), NSCompositeSourceOver);				}
				[image drawInRect:dstRect fromRect:NSZeroRect operation:NSCompositeSourceOver fraction:fraction];
			}
		}
	}
	
	if (draggingKomaPosition.x >= 0 && draggingKomaPosition.y >= 0) {
		dstRect.origin.x = draggingKomaPosition.x - kKomaXSize * 0.5;
		dstRect.origin.y = draggingKomaPosition.y - kKomaYSize * 0.5;
		dstRect.size.width = kKomaXSize;
		dstRect.size.height = kKomaYSize;
		image = images[moving.bw][moving.koma];
		[image drawInRect:dstRect fromRect:NSZeroRect operation:NSCompositeSourceOver fraction:1.0];
	}
	
	for (i = 0; i < 2; i++) {
		id tebanInfo = (i == 0 ? blackTebanInfo : whiteTebanInfo);
		dstRect.origin = [tebanInfo bounds].origin;
		dstRect.origin = [boardView convertPoint:[tebanInfo convertPoint:dstRect.origin toView:nil] fromView:nil];
		dstRect.origin.x -= kKomaXSize * 0.8;
		dstRect.origin.y -= kKomaYSize * 0.3;
		dstRect.size = NSMakeSize(kKomaXSize * 0.6, kKomaYSize * 0.6);
		[(i == 0 ? gBlackFront : gWhiteFront) drawInRect:dstRect fromRect:NSZeroRect operation:NSCompositeSourceOver fraction:fraction];
	}

	[boardView unlockFocus];
}

- (void)updateTebanInfo
{
	int i;
	NSString *blackPlayer = (handicap > 0 ? NSLocalizedString(@"Handicapped black", @"shitate") : NSLocalizedString(@"Black", @"black"));
	NSString *whitePlayer = (handicap > 0 ? NSLocalizedString(@"Handicapped white", @"uwate") : NSLocalizedString(@"White", @"white"));
	NSString *yourName = NSLocalizedString(@"You", @"you");
	NSString *bonanzaName = NSLocalizedString(@"Bonanza", @"bonanza");
	[blackName setStringValue:[NSString stringWithFormat:@"%@: %@", blackPlayer, (isHuman[0] ? yourName : bonanzaName)]];
	[whiteName setStringValue:[NSString stringWithFormat:@"%@: %@", whitePlayer, (isHuman[1] ? yourName : bonanzaName)]];
	if (status >= 1 && status <= 4) {
		NSString *resignInfo;
		if (status == 1 || status == 2)
			resignInfo = NSLocalizedString(@"Resigned", @"resigned");
		else
			resignInfo = NSLocalizedString(@"Mated", @"mated");
		if (status == 1 || status == 3) {
			[blackTebanInfo setStringValue:resignInfo];
			[blackTebanInfo setHidden:NO];
			[blackTebanInfo setBackgroundColor:[NSColor redColor]];
			[whiteTebanInfo setHidden:YES];
		} else {
			[whiteTebanInfo setStringValue:resignInfo];
			[whiteTebanInfo setHidden:NO];
			[whiteTebanInfo setBackgroundColor:[NSColor redColor]];
			[blackTebanInfo setHidden:YES];
		}
	} else {
		[blackTebanInfo setHidden:(teban == 0 ? NO : YES)];
		[whiteTebanInfo setHidden:(teban == 1 ? NO : YES)];
		[(teban == 0 ? blackTebanInfo : whiteTebanInfo) setBackgroundColor:[NSColor whiteColor]];
	}
	[whiteChecked setHidden:(status == 3 || status == 5 ? NO : YES)];
	[blackChecked setHidden:(status == 4 || status == 6 ? NO : YES)];
	for (i = 0; i < 2; i++) {
		int n = elapsedTime[i];
		NSString *s = [NSString stringWithFormat:@"%02d:%02d:%02d", n / 3600, (n % 3600) / 60, n % 60];
		[(i == 0 ? blackElapsedTime : whiteElapsedTime) setStringValue:s];
	}
	[undoButton setEnabled:[self canUndo]];		
}

- (void)moveKomaByMouse:(NSPoint)pos
{
	int x, y;
	x = pos.x - kKomaXOrigin;
	y = pos.y - kKomaYOrigin;
	if (x >= 0 && x < kKomaXSize * 9 && y >= 0 && y < kKomaYSize * 9) {
		x = 8 - x / kKomaXSize;
		y = 8 - y / kKomaYSize;
		if (board_focus[x + y * 9] == 1) {
			moving.pos[2] = x + 1;
			moving.pos[3] = y + 1;
			moving.promote = 0;
			moving.moveTime = elapsedTime[teban] - lastMoveTime;
			[self stopTimer];
			if (moving.pos[0] > 0 && moving.pos[1] > 0) {
				if (moving.koma < 10 && moving.koma != 5 && moving.koma != 8) {
					/*  Ask for promotion  */
					int y1, y2;
					BOOL promote = NO;
					y1 = (moving.bw == 0 ? moving.pos[1] : 10 - moving.pos[1]);
					y2 = (moving.bw == 0 ? moving.pos[3] : 10 - moving.pos[3]);
					if (y1 <= 3 || y2 <= 3) {
						if ((moving.koma == 1 || moving.koma == 2) && y2 == 1)
							promote = YES;
						else if (moving.koma == 3 && y2 <= 2)
							promote = YES;
						else {
							int modalStatus;
							NSPoint balloonPos = NSMakePoint(kKomaXOrigin + (9 - x) * kKomaXSize, kKomaYOrigin + (8 - y) * kKomaYSize);
							balloonPos = [window convertBaseToScreen:[shogiView convertPoint:balloonPos toView:nil]];
							draggingKomaPosition.x = kKomaXOrigin + (8 - x) * kKomaXSize + kKomaXSize / 2;
							draggingKomaPosition.y = kKomaYOrigin + (8 - y) * kKomaYSize + kKomaYSize / 2;
							[shogiView display];
							[promoteWindow makeKeyAndOrderFront:self];
							[promoteWindow setFrameOrigin:balloonPos];
							modalStatus = [NSApp runModalForWindow:promoteWindow];
							if (modalStatus == NSRunStoppedResponse)
								promote = YES;
							[promoteWindow orderOut:self];
						}
						if (promote) {
							moving.koma += 10;
							moving.promote = 1;
						}
					}
				}
			}

			/*  Captured koma  */
			moving.capture = board[x + y * 9];
			if ((signed char)moving.capture < 0)
				moving.capture = -moving.capture;

			/*  Notify bonanza  */
			if (teban == moving.bw) {
				if (moving.koma == 99) {
					[self writeStringToBon:@"resign\n"];
				} else {
					int i;
					unsigned char p[4];
					for (i = 0; i < 4; i++) {
						if (moving.pos[i] == 0)
							p[i] = 0;
						else
							p[i] = (whiteMovesFirst ? 10 - moving.pos[i] : moving.pos[i]);
					}
					[self writeStringToBon:[NSString stringWithFormat:@"%d%d%d%d%s\n", p[0], p[1], p[2], p[3], komaCodeToString(moving.koma)]];
				}
			}
			/*  Move on the display  */
			[self makeMove:&moving withAnimation:NO];
			/*  Restart timer  */
			[self startTimer];
		}
	}
	moving.koma = 0;
}

- (void)makeMove:(const BonMoveInfo *)iptr withAnimation:(BOOL)showAnimation
{
	NSPoint pos1, pos2;
	int i, k, koma;

	if (iptr->koma == 99 || iptr->koma == 98) {
		/*  resign or mate */
		status = iptr->bw + (iptr->koma == 99 ? 1 : 3);
		teban = -1;
	} else {
		NSRect r;
		k = iptr->pos[2] + iptr->pos[3] * 9 - 10;
		if (iptr->pos[0] == 0 && iptr->pos[1] == 0) {
			/*  Drop  */
			if (showAnimation) {
				r = mochigomaRect(iptr->bw, iptr->koma, 0, mochigoma[iptr->bw][iptr->koma - 1], whiteIsThisSide);
				pos1.x = r.origin.x + r.size.width * 0.5;
				pos1.y = r.origin.y + r.size.height * 0.5;
			}
			mochigoma[iptr->bw][iptr->koma - 1]--;
		} else {
			/*  Move  */
			board[iptr->pos[0] + iptr->pos[1] * 9 - 10] = 0;
			if (showAnimation) {
				pos1.x = (whiteIsThisSide ? iptr->pos[0] - 1 : 9 - iptr->pos[0]);
				pos1.y = (whiteIsThisSide ? iptr->pos[1] - 1 : 9 - iptr->pos[1]);
				pos1.x = kKomaXOrigin + kKomaXSize * (pos1.x + 0.5);
				pos1.y = kKomaYOrigin + kKomaYSize * (pos1.y + 0.5);
			}
		}
		if (showAnimation) {
			BonMoveInfo saveMoving = moving;
			moving = *iptr;
			pos2.x = (whiteIsThisSide ? iptr->pos[2] - 1 : 9 - iptr->pos[2]);
			pos2.y = (whiteIsThisSide ? iptr->pos[3] - 1 : 9 - iptr->pos[3]);
			pos2.x = kKomaXOrigin + kKomaXSize * (pos2.x + 0.5);
			pos2.y = kKomaYOrigin + kKomaYSize * (pos2.y + 0.5);
			r.origin.x = draggingKomaPosition.x - kKomaXSize * 0.5;
			r.origin.y = draggingKomaPosition.y - kKomaYSize * 0.5;
			r.size = NSMakeSize(kKomaXSize, kKomaYSize);
			[shogiView display];
			for (i = 1; i <= 30; i++) {
				[shogiView setNeedsDisplayInRect:r];
				draggingKomaPosition.x = floor(pos1.x + i * (pos2.x - pos1.x) / 30.0);
				draggingKomaPosition.y = floor(pos1.y + i * (pos2.y - pos1.y) / 30.0);
				r.origin.x = draggingKomaPosition.x - kKomaXSize * 0.5;
				r.origin.y = draggingKomaPosition.y - kKomaYSize * 0.5;
				[shogiView setNeedsDisplayInRect:r];
				[shogiView displayIfNeeded];
				usleep(15000);
			}
			draggingKomaPosition = NSMakePoint(-1, -1);
			moving = saveMoving;
		}
		if (board[k] != 0) {
			/*  Capture  */
			koma = board[k];
			if (koma != 0) {
				if (koma < 0)
					koma = -koma;
				if (koma >= 10)
					koma -= 10;
				if (koma >= 1 && koma <= 7)
					mochigoma[iptr->bw][koma - 1]++;
			}
		}
		board[k] = iptr->koma * (iptr->bw == 0 ? 1 : -1);
		teban = 1 - teban;
		if ([self isChecked:teban]) {
			if ([self isMated:teban]) {
				status = teban + 3;
				teban = -1;
			} else status = teban + 5;
		} else status = 0;
	}

	/*  Log moves  */
	if (bonMoveIndex < [bonMoves count]) {
		for (i = [bonMoves count] - 1; i >= bonMoveIndex; i--) {
			[bonMoves removeObjectAtIndex:i];
		}
	}
	[bonMoves addObject:[NSData dataWithBytes:iptr length:sizeof(*iptr)]];
	bonMoveIndex++;
	
	[self updateTebanInfo];
	[shogiView setNeedsDisplay:YES];
	
	if (status >= 1 && status <= 4) {
		/*  Game Over  */
		int winPlayer = status % 2;  /* 0: black, 1: white */
		NSString *winString;
		if (isHuman[0] == isHuman[1]) {
			/*  "Black" (Sente, Uwate) or "White" (Gote, Shitate)  */
			if (handicap > 0)
				winString = (winPlayer ? NSLocalizedString(@"Handicapped white", @"shitate") : NSLocalizedString(@"Handicapped black", @"uwate"));
			else
				winString = (winPlayer ? NSLocalizedString(@"White", @"white") : NSLocalizedString(@"Black", @"black"));
		} else {
			if (isHuman[winPlayer])
				winString = NSLocalizedString(@"You", @"You");
			else
				winString = NSLocalizedString(@"Bonanza", @"Bonanza");
		}
		winString = [NSString stringWithFormat:NSLocalizedString(@"Who win", @"%@ win."), winString];
		[shogiView display];
		usleep(1000000);
		[winText setStringValue:winString];
		[NSApp beginSheet:gameOverWindow modalForWindow:window modalDelegate:nil didEndSelector:nil contextInfo:nil];
	}
}

- (void)makeReverseMove:(const BonMoveInfo *)iptr withAnimation:(BOOL)showAnimation
{
	NSPoint pos1, pos2;
	int i, k, sign;
	NSRect r;
	k = iptr->pos[2] + iptr->pos[3] * 9 - 10;
	sign = (iptr->bw == 0 ? 1 : -1);
	if (iptr->pos[0] == 0 && iptr->pos[1] == 0) {
		/*  Undo drop  */
		if (showAnimation) {
			r = mochigomaRect(iptr->bw, iptr->koma, 0, mochigoma[iptr->bw][iptr->koma - 1] + 1, whiteIsThisSide);
			pos1.x = r.origin.x + r.size.width * 0.5;
			pos1.y = r.origin.y + r.size.height * 0.5;
		}
		mochigoma[iptr->bw][iptr->koma - 1]++;
		board[k] = 0;
	} else {
		/*  Undo move  */
		if (showAnimation) {
			pos1.x = (whiteIsThisSide ? iptr->pos[0] - 1 : 9 - iptr->pos[0]);
			pos1.y = (whiteIsThisSide ? iptr->pos[1] - 1 : 9 - iptr->pos[1]);
			pos1.x = kKomaXOrigin + kKomaXSize * (pos1.x + 0.5);
			pos1.y = kKomaYOrigin + kKomaYSize * (pos1.y + 0.5);
		}
		if (iptr->capture != 0) {
			board[k] = -sign * iptr->capture;
			mochigoma[iptr->bw][iptr->capture % 10 - 1]--;
		} else {
			board[k] = 0;
		}
	}
	if (showAnimation) {
		BonMoveInfo saveMoving = moving;
		moving = *iptr;
		pos2.x = (whiteIsThisSide ? iptr->pos[2] - 1 : 9 - iptr->pos[2]);
		pos2.y = (whiteIsThisSide ? iptr->pos[3] - 1 : 9 - iptr->pos[3]);
		pos2.x = kKomaXOrigin + kKomaXSize * (pos2.x + 0.5);
		pos2.y = kKomaYOrigin + kKomaYSize * (pos2.y + 0.5);
		r.origin.x = draggingKomaPosition.x - kKomaXSize * 0.5;
		r.origin.y = draggingKomaPosition.y - kKomaYSize * 0.5;
		r.size = NSMakeSize(kKomaXSize, kKomaYSize);
		[shogiView display];
		for (i = 1; i <= 30; i++) {
			[shogiView setNeedsDisplayInRect:r];
			draggingKomaPosition.x = floor(pos2.x + i * (pos1.x - pos2.x) / 30.0);
			draggingKomaPosition.y = floor(pos2.y + i * (pos1.y - pos2.y) / 30.0);
			r.origin.x = draggingKomaPosition.x - kKomaXSize * 0.5;
			r.origin.y = draggingKomaPosition.y - kKomaYSize * 0.5;
			[shogiView setNeedsDisplayInRect:r];
			[shogiView displayIfNeeded];
			usleep(15000);
		}
		draggingKomaPosition = NSMakePoint(-1, -1);
		moving = saveMoving;
	}
	if (iptr->pos[0] > 0 && iptr->pos[1] > 0) {
		board[iptr->pos[0] + iptr->pos[1] * 9 - 10] = sign * (iptr->promote ? iptr->koma - 10 : iptr->koma);
	}
	[shogiView setNeedsDisplay:YES];
}

#pragma mark ====== Mouse Actions ======

- (void)mouseDown:(NSEvent *)theEvent inView:(NSView *)boardView
{
	NSPoint pos = [boardView convertPoint:[theEvent locationInWindow] fromView:nil];
	int x, y, k;
	if (moving.koma > 0)
		return;  /*  Already moving: wait for mouseUp event  */
	memset(board_focus, 0, 81);
	memset(&moving, 0, sizeof(moving));
	x = pos.x - kKomaXOrigin;
	y = pos.y - kKomaYOrigin;
	if (x >= 0 && x < kKomaXSize * 9 && y >= 0 && y < kKomaYSize * 9) {
		x = 8 - x / kKomaXSize;
		y = 8 - y / kKomaYSize;
		k = board[x + y * 9];
		if (k != 0) {
			if (teban >= 0 && ((teban == 0 && k < 0) || (teban == 1 && k > 0)))
				return;
			[self setAllowablePositions:board_focus forKomaAtX:x + 1 y:y + 1];
			board_focus[x + y * 9] = 3;
			moving.bw = (k < 0 ? 1 : 0);
			moving.koma = (k < 0 ? -k : k);
			moving.pos[0] = x + 1;
			moving.pos[1] = y + 1;
		} else return;  /*  No operation  */
		draggingKomaPosition.x = kKomaXOrigin + (8 - x) * kKomaXSize + kKomaXSize / 2;
		draggingKomaPosition.y = kKomaYOrigin + (8 - y) * kKomaYSize + kKomaYSize / 2;
	} else {
		for (k = 1; k < 8; k++) {
			for (moving.bw = 0; moving.bw < 2; moving.bw++) {
				int n, nn = mochigoma[moving.bw][k - 1];
				if (teban >= 0 && moving.bw != teban)
					continue;
				for (n = 0; n < nn; n++) {
					NSRect rect = mochigomaRect(moving.bw, k, n, nn, whiteIsThisSide);
					if (NSMouseInRect(pos, rect, [boardView isFlipped])) {
						moving.koma = k;
						draggingMochigomaIndex = n;
						draggingKomaPosition.x = rect.origin.x + rect.size.width / 2;
						draggingKomaPosition.y = rect.origin.y + rect.size.height / 2;
						goto end_loop;
					}
				}
			}
		}
	end_loop:
		if (moving.koma != 0) {
			[self setDroppablePositions:board_focus forKoma:(moving.koma * (moving.bw == 0 ? 1 : -1))];
		} else return;  /*  No operation  */
	}
	isDraggingKoma = NO;
	[boardView display];
}

- (void)mouseDragged:(NSEvent *)theEvent inView:(NSView *)boardView
{
	NSRect r;
	NSPoint pos = [boardView convertPoint:[theEvent locationInWindow] fromView:nil];
	if (moving.koma == 0)
		return;  /*  No dragging  */
	if (!isDraggingKoma && [theEvent type] == NSLeftMouseDragged) {
		/*  If drag amount is small then don't start dragging yet  */
		if (fabs(pos.x - draggingKomaPosition.x) < 2 && fabs(pos.y - draggingKomaPosition.y) < 2)
			return;
	}
	isDraggingKoma = YES;
	r.origin.x = draggingKomaPosition.x - kKomaXSize * 0.5;
	r.origin.y = draggingKomaPosition.y - kKomaYSize * 0.5;
	r.size = NSMakeSize(kKomaXSize, kKomaYSize);
	[boardView setNeedsDisplayInRect:r];
	draggingKomaPosition = pos;
	r.origin.x = draggingKomaPosition.x - kKomaXSize * 0.5;
	r.origin.y = draggingKomaPosition.y - kKomaYSize * 0.5;
	[boardView setNeedsDisplayInRect:r];
}

- (void)mouseMoved:(NSEvent *)theEvent inView:(NSView *)boardView
{
	if (moving.koma != 0)
		[self mouseDragged:theEvent inView:boardView];
}

- (void)mouseUp:(NSEvent *)theEvent inView:(NSView *)boardView
{
	NSPoint pos = [boardView convertPoint:[theEvent locationInWindow] fromView:nil];
	if (moving.koma == 0)
		return;  /*  No dragging  */
	if (isDraggingKoma) {
		/*  Move koma  */
		[self moveKomaByMouse:pos];
	} else if ([theEvent clickCount] > 0) {
		/*  Start moving koma by click  */
		isDraggingKoma = YES;
		[window setAcceptsMouseMovedEvents:YES];
	} else {
		/*  No action: clear moving koma  */
		moving.koma = 0;
	}
	if (moving.koma == 0) {
		/*  End moving  */
		isDraggingKoma = NO;
		memset(&moving, 0, sizeof(moving));
		memset(board_focus, 0, 81);
		draggingKomaPosition = NSMakePoint(-1, -1);
		[window setAcceptsMouseMovedEvents:NO];
	}
	[shogiView setNeedsDisplay:YES];
}

#pragma mark ====== Time ======

- (void)startTimer
{
	if (timer != nil)
		[self stopTimer];
	timer = [[NSTimer scheduledTimerWithTimeInterval:1.0 target:self selector:@selector(timerFired:) userInfo:nil repeats:YES] retain];
	lastMoveTime = elapsedTime[teban];
}

- (void)stopTimer
{
	if (timer != nil) {
		[timer invalidate];
		[timer autorelease];
		timer = nil;
	}
}

- (void)timerFired:(NSTimer *)aTimer
{
	if (teban == 0 || teban == 1) {
		elapsedTime[teban]++;
		[self updateTebanInfo];
	}
}

#pragma mark ====== Actions ======

- (IBAction)okPressed:(id)sender
{
	[NSApp stopModal];
}

- (IBAction)cancelPressed:(id)sender
{
	[NSApp abortModal];
}

- (IBAction)undoMove:(id)sender
{
	const BonMoveInfo *iptr, *iptr2;
	int i;
	if (![self canUndo])
		return;

	if (isHuman[0] == 0 || isHuman[1] == 0)
		[self stopBonanza];

	iptr = (const BonMoveInfo *)[[bonMoves objectAtIndex:--bonMoveIndex] bytes];
	[self makeReverseMove:iptr withAnimation:YES];
	teban = 1 - teban;
	if (isHuman[iptr->bw] == 0) {
		/*  This is for computer move: undo another move  */
		iptr2 = (const BonMoveInfo *)[[bonMoves objectAtIndex:--bonMoveIndex] bytes];
		[self makeReverseMove:iptr2 withAnimation:YES];
		teban = 1 - teban;
	} else iptr2 = NULL;
	if (isHuman[0] == 0 || isHuman[1] == 0) {
		/*  Restart game for bonanza  */
		[self startBonanza];
		[self writeStringToBon:bonanzaInitString];
		for (i = 0; i < bonMoveIndex; i++) {
			int j;
			unsigned char p[4];
			iptr = (const BonMoveInfo *)[[bonMoves objectAtIndex:i] bytes];
			for (j = 0; j < 4; j++) {
				if (iptr->pos[j] == 0)
					p[j] = 0;
				else
					p[j] = (whiteMovesFirst ? 10 - iptr->pos[j] : iptr->pos[j]);
			}
			[self writeStringToBon:[NSString stringWithFormat:@"move %d%d%d%d%s\n", p[0], p[1], p[2], p[3], komaCodeToString(iptr->koma)]];
		}
	}
	[self updateTebanInfo];
}

- (IBAction)resignGame:(id)sender
{
	if (isHuman[teban]) {
		[self writeStringToBon:@"resign\n"];
		moving.koma = 99;
		moving.bw = teban;
		[self makeMove:&moving withAnimation:NO];
		teban = -1;
	}
}

- (IBAction)startNewGame:(id)sender
{
//	NSArray *array = NSSearchPathForDirectoriesInDomains(NSApplicationSupportDirectory, NSUserDomainMask, YES);
//	NSString *bonPath = [[[array objectAtIndex:0] stringByAppendingPathComponent:@"BonanzaX"] retain];

	[self setUpNewGame];

	teban = (whiteMovesFirst ? 1 : 0);
	
	if (isHuman[0] == 0 || isHuman[1] == 0) {
		
		/*  Initialize bonanza setup  */
		NSUserDefaults *def = [NSUserDefaults standardUserDefaults];
		int time1, time2, resignLimit;

		[self startBonanza];

		[bonanzaInitString release];
		time1 = [[def objectForKey:gAvailableTime1] intValue];
		time2 = [[def objectForKey:gAvailableTime2] intValue];
		if (handicap == 0)
			resignLimit = 0;
		else if (handicap < 8)
			resignLimit = 10000;
		else resignLimit = 100000;
		bonanzaInitString = [NSString stringWithFormat:@"limit time %d %d\nnew PI%@\n", time1, time2, handicapSpec];
		if (resignLimit > 0)
			bonanzaInitString = [bonanzaInitString stringByAppendingFormat:@"resign %d\n", resignLimit];
		[self writeStringToBon:bonanzaInitString];
		[bonanzaInitString retain];

		/*  Ask for move if bonanza is first to move  */
		if (isHuman[teban] == 0)
			[self writeStringToBon:@"move\n"];
	}
	
	elapsedTime[0] = elapsedTime[1] = 0;
	[self startTimer];

	[shogiView setNeedsDisplay:YES];
	[self updateTebanInfo];
}

- (IBAction)openSetUpSheet:(id)sender
{
	[NSApp beginSheet:settingsWindow modalForWindow:window modalDelegate:nil didEndSelector:nil contextInfo:nil];
}

- (IBAction)okSetUpSheet:(id)sender
{
	[NSApp endSheet:settingsWindow];
	[settingsWindow orderOut:self];
	[self setUpNewGame];
}

- (IBAction)cancelSetUpSheet:(id)sender
{
	[NSApp endSheet:settingsWindow];
	[settingsWindow orderOut:self];
}

- (IBAction)okGameOverSheet:(id)sender
{
	[NSApp endSheet:gameOverWindow];
	[gameOverWindow orderOut:self];
}

- (IBAction)okAboutWindow:(id)sender
{
	[NSApp endSheet:aboutWindow];
	[aboutWindow orderOut:self];
}

- (IBAction)showAboutWindow:(id)sender
{
	static NSString *sLastBuildDate;
	NSString *versionString;
	if (sLastBuildDate == nil) {
		NSString *path = [[NSBundle mainBundle] pathForResource:@"last_build" ofType:@"txt"];
		FILE *fp = fopen([path fileSystemRepresentation], "r");
		sLastBuildDate = @"";
		if (fp != NULL) {
			char buf[128];
			if (fgets(buf, sizeof buf, fp) != NULL) {
				char *p;
				for (p = buf; *p != 0 && *p != '\n'; p++);
				*p = 0;
				sLastBuildDate = [[NSString stringWithUTF8String:buf] retain];
			}
		}
	}
	versionString = [[[NSBundle mainBundle] infoDictionary] objectForKey:@"CFBundleShortVersionString"];
	if (versionString == nil)
		versionString = [[[NSBundle mainBundle] infoDictionary] objectForKey:@"CFBundleVersion"];
	if (versionString == nil)
		versionString = @"NA";
	[aboutVersionText setStringValue:[NSString stringWithFormat:NSLocalizedString(@"Version %@", @"Version (about window)"), versionString]];
	[aboutBuildDateText setStringValue:[NSString stringWithFormat:NSLocalizedString(@"Last Build: %@", @"Last Build (about window)"), sLastBuildDate]];
	[NSApp beginSheet:aboutWindow modalForWindow:window modalDelegate:nil didEndSelector:nil contextInfo:nil];
	
}

#pragma mark ====== NSApplication Overrides ======

- (void)applicationWillFinishLaunching:(NSNotification *)notification
{
	NSUserDefaults *def = [NSUserDefaults standardUserDefaults];
	int i;
	/*  Load default settings  */
	NSString *initialValues[] = {
		gKomaImage, @"0", gBanImage, @"0", gGridImage, @"0",
		gBlackIsHuman, @"1", gWhiteIsHuman, @"0", gHandicap, @"0",
		gAvailableTime1, @"0", gAvailableTime2, @"1", nil };
	
	for (i = 0; initialValues[i] != nil; i += 2) {
		id obj;
		obj = [def objectForKey:initialValues[i]];
		if (obj == nil)
			[def setObject:initialValues[i + 1] forKey:initialValues[i]];
	}
	
}

- (void)applicationDidFinishLaunching:(NSNotification *)aNotification
{
	NSUserDefaults *def = [NSUserDefaults standardUserDefaults];
	
	/*  Prepare temporary directory  */
	NSArray *array = NSSearchPathForDirectoriesInDomains(NSApplicationSupportDirectory, NSUserDomainMask, YES);
	bonPath = [[[array objectAtIndex:0] stringByAppendingPathComponent:@"BonanzaX"] retain];
	system([[NSString stringWithFormat:@"mkdir -p '%@/log'", bonPath] fileSystemRepresentation]);
	
	/*  Create symbolic links  */
	/*  fv.bin is read-only. Other files should be read/write accessible  */
	NSString *bookPath = [[NSBundle mainBundle] pathForResource:@"bonanza_book" ofType:nil];
	system([[NSString stringWithFormat:@"cd '%@'; rm -f book.bin fv.bin hash.bin flip.csa", bonPath] fileSystemRepresentation]);	
	system([[NSString stringWithFormat:@"cd '%@'; cp -n '%@/book.bin' book.bin", bonPath, bookPath] fileSystemRepresentation]);
	system([[NSString stringWithFormat:@"cd '%@'; ln -sf '%@/fv.bin' fv.bin", bonPath, bookPath] fileSystemRepresentation]);
	system([[NSString stringWithFormat:@"cd '%@'; cp -n '%@/hash.bin' hash.bin", bonPath, bookPath] fileSystemRepresentation]);
	system([[NSString stringWithFormat:@"cd '%@'; cp -n '%@/flip.csa' flip.csa", bonPath, bookPath] fileSystemRepresentation]);
	
	bonMoves = [[NSMutableArray alloc] init];
	
	[self loadKomaImages:[[def objectForKey:gKomaImage] intValue]];
	[self loadBanImage:[[def objectForKey:gBanImage] intValue]];
	[self loadGridImage:[[def objectForKey:gGridImage] intValue]];
	
	/*  Make transparent window for "Promote?" dialog  */
	[promoteWindow setBackgroundColor:[NSColor clearColor]];
	[promoteWindow setOpaque:NO];
	
	draggingKomaPosition = NSMakePoint(-1, -1);
	teban = -1;
	
	[window makeKeyAndOrderFront:self];
	
	[self setUpNewGame];
}

- (void)applicationWillTerminate:(NSNotification *)aNotification
{
	if (bonTask != nil) {
		if ([bonTask isRunning])
			[bonTask terminate];
		[bonTask release];
	}
	if (pipeToBon != nil)
		[pipeToBon release];
	if (pipeFromBon != nil)
		[pipeFromBon release];
//	system([[NSString stringWithFormat:@"cd '%@'; rm book.bin fv.bin hash.bin flip.csa", bonPath] fileSystemRepresentation]);
}

@end
